parameters = {
    "global": {
        "app-name": "my-app",
        "app-key": "my-app-key",
        "event-type": "scheduler",
        "query-limit": 1000,
    },
}


def get(key, default=None):
    if not key:
        if default:
            return default

        raise KeyError("No key provided")

    path = key.split(".")
    value = parameters

    while path:
        current_key = path.pop(0)

        if current_key not in value:
            if default:
                return default

            raise KeyError("{0} not found in path".format(current_key))

        value = value.get(current_key)

    return value


def update(additional_parameters):
    """
    The purpose of this method is to update the existing parameters data
    with the provided additional parameters. Note that the globals will
    be replaced if the additional parameters contains global node.
    :param additional_parameters:
    :return:
    """
    globals = parameters.pop('global')
    additional_globals = additional_parameters.pop('global', {})
    globals.update(additional_globals)
    parameters.update(additional_parameters)
    if globals:
        parameters['global'] = globals



import json

from worker.framework.mixins import RedisMixin


def file_to_json(file_name):
    with open(file_name, mode='r') as file:
        _json = json.load(file)
        return _json


def get_last_processed_timestamp(redis_key: str):
    """
    Get the last_processed_timestamp from the redis key
    :param redis_key: redis key
    """
    try:
        r = RedisMixin().get_redis()
        previous_state = r.get(redis_key)
        previous_state = json.loads(previous_state)
        return previous_state.get('last_processed_timestamp', None)
    except Exception as ex:
        print(f"REDIS connection error! Error: {ex}")

    return None


def create_scheduler_events(asset_id, start_timestamp, end_timestamp, step):
    """
    Creating scheduler events
    :param asset_id:
    :param start_timestamp:
    :param end_timestamp:
    :param step:
    :return:
    """
    triggers = range(start_timestamp, end_timestamp, step)
    events = [[[{"asset_id": asset_id, "schedule_start": 1000 * t}]] for t in triggers]
    return events



import os

import redis

from worker.data.operations import is_number


class RedisMixin(object):
    def __init__(self, *args, **kwargs):
        pass

    def get_redis(self):
        if not getattr(self, 'redis', None):
            key = os.getenv('CACHE_URL', None)

            if not key:
                raise Exception('redis key (CACHE_URL) not found in Environment Variables.')

            self.redis = redis.Redis.from_url(key)

        return self.redis

    @staticmethod
    def get_redis_key(asset_id, app_name, module_name=None):
        key = "corva/{0}.{1}".format(asset_id, app_name)
        if module_name:
            key += ".{0}".format(module_name)

        return key

    @classmethod
    def delete_states(cls, pattern: str):
        """
        Delete the states from redis that matches the pattern
        :param pattern: pattern
        :return:
        """
        r = cls().get_redis()
        if not r:
            raise Exception('redis connection not established')

        redis_state_keys = r.keys(pattern)
        for key in redis_state_keys:
            r.delete(key)


class LoggingMixin(object):

    LOGGING_LEVELS = [
        "all",  # lowest level
        "debug",
        "info",
        "warn",
        "error",
        "fatal",
        "off"
    ]

    def init_logging(self, *args, **kwargs):
        self.logging_level = os.getenv("LOGGING_LEVEL", "info").lower()
        self.logging_asset_id = os.getenv("LOGGING_ASSET_ID", 0)

        if self.logging_level not in self.LOGGING_LEVELS:
            self.logging_level = "off"

        if is_number(self.logging_asset_id):
            self.logging_asset_id = int(self.logging_asset_id)

    def all(self, asset_id, text):
        self.call('all', asset_id, text)

    def debug(self, asset_id, text):
        self.call('debug', asset_id, text)

    def info(self, asset_id, text):
        self.call('info', asset_id, text)

    def warn(self, asset_id, text):
        self.call('warn', asset_id, text)

    def error(self, asset_id, text):
        self.call('error', asset_id, text)

    def fatal(self, asset_id, text):
        self.call('fatal', asset_id, text)

    def off(self, asset_id, text):
        pass

    def call(self, method, asset_id, text):
        if str(method) in self.LOGGING_LEVELS[0:self.LOGGING_LEVELS.index(self.logging_level) + 1] \
                and asset_id == self.logging_asset_id:
            self.log(asset_id, text)

    @staticmethod
    def log(asset_id, text):
        print('asset_id_{} -> {}'.format(asset_id, text))


class RollbarMixin(object):
    def init_rollbar(self, *args, **kwargs):
        self.rollbar = kwargs.pop('rollbar', None)
        super().__init__(*args, **kwargs)

    def track_error(self, message: str, level: str):
        # Levels: critical, error, warning, info, debug, and ignored
        if not self.rollbar:
            return

        level = level.lower()

        self.rollbar.report_message(message, level)



import numpy as np

from worker.data.api import API


def gather_wits_for_period(asset_id, start, end, limit=1800):
    query = '{timestamp#gte#%s}AND{timestamp#lte#%s}' % (start, end)
    worker = API()
    wits_dataset = worker.get(
        path="/v1/data/corva", collection='wits', asset_id=asset_id, sort="{timestamp: 1}", limit=limit, query=query
    ).data

    if not wits_dataset:
        return None

    return wits_dataset


def get_one_wits_record(asset_id, timestamp_sort=-1):
    api_worker = API()
    data = api_worker.get(path='/v1/data/corva/', collection='wits', asset_id=asset_id,
                          sort="{timestamp:%s}" % timestamp_sort, limit=1).data
    if not data:
        return None

    timestamp = data[0].get('timestamp', 0)
    return timestamp


def get_config_by_id(config_collection, _id):
    worker = API()
    config = worker.get(path="/v1/data/corva/%s/%s" % (config_collection, _id)).data
    return config


def delete_collection_data_of_asset_id(asset_id, collection):
    """
    Delete all the data of a collection for an asset id.
    :param asset_id:
    :param collection:
    :return:
    """
    worker = API()

    path = "/v1/data/corva/%s" % collection
    query = "{asset_id#eq#%s}" % asset_id
    worker.delete(path=path, query=query)
    print(f"Deleted {collection} for asset_id {asset_id}.")


def is_number(data):
    """
    Check and return True if data is a number, else return False
    :param data: Input can be string, number or nan
    :return: True or False
    """
    try:
        data_cast = float(data)
        if data_cast >= 0 or data_cast <= 0:  # to make sure it is a valid number
            return True

        return False
    except ValueError:
        return False
    except TypeError:
        return False


def is_finite(data):
    """
    Check if the given data is a finite number
    :param data: Input can be string, number or nan
    :return: True or False
    """
    return is_number(data) and np.isfinite(data)


def to_number(data):
    """
    Check and return if the data can be cast to a number, else return None
    :param data: Input can be string, number or nan
    :return: A numbers
    """
    if is_number(data):
        return float(data)

    return None


def none_to_nan(data):
    """
    If data is a list, return list with None replaced with nan.
    If data is None, return nan
    :param data:
    :return:
    """
    if isinstance(data, list):
        return [np.nan if e is None else e for e in data]

    if data is None:
        return np.nan

    return data


def get_value_from_dict(data: dict, key: str, func=lambda x: x, default=None):
    """
    An structured way of getting data from a dict.
    :param data:
    :param key:
    :param func: the type of the data (int, str, float, ...)
    :param default:
    :return:
    """
    value = data.get(key, None)
    if value is not None:
        try:
            return_value = func(value)
            return return_value
        except ValueError:
            pass

    return default


def get_dict_data_by_path(data: dict, key_path: str, default=None):
    """
    To find the path to a key in data dictionary.
    Note that none of the keys should end up in a list
    :param data:
    :param key_path: path to the final key, e.g. 'data.X.Y'
    :param default:
    :return:
    """
    keys = key_path.split('.')
    for key in keys:
        data = data.get(key, {})
    if not data:
        return default
    return data


def is_in_and_not_none(d: dict, key: str):
    """
    An structured way of getting data from a dict.
    :param d: the dictionary
    :param key:
    :return: True or False
    """
    if key in d.keys() and d[key] is not None:
        return True

    return False


def nanround(value, decimal_places=2):
    """
    Similar to python built-in round but considering None values as well
    :param value:
    :param decimal_places:
    :return:
    """
    if is_number(value):
        return round(value, decimal_places)

    return None


def merge_dicts(d1: dict, d2: dict) -> dict:
    """
    Merge two dictionaries
    Note: the 2nd item (d2) has a higher priority to write items with similar keys
    :param d1:
    :param d2:
    :return:
    """
    d = {**d1, **d2}
    return d



# flake8: noqa: F401


from worker.data.api import API
from worker.framework import constants
from worker.app import App
from worker.app.modules import Module
from worker.test.lambda_function_test_run import AppTestRun



"""
This file is used to trigger lambda functions locally and test
the results of the app on actual data to make sure of the results.
An example to use this functionality is like the one shows on
'app_test.py' and then that file can run in command line.

Note: make sure you run your tests on QA environment.

== app_test.py file
# added the main directory as the first path
from os.path import dirname
import sys
parent = dirname(dirname(__file__))
sys.path.insert(0, parent)

if __name__ == '__main__':
    collections = ['collection_to_delete']
    app = AppTestRun(lambda_function.lambda_handler, collections)
    app.run()

> python app_test -a 16886 -d True
"""

import argparse
from distutils.util import strtobool
from tqdm import tqdm

from worker import constants
from worker.data.operations import delete_collection_data_of_asset_id, get_one_wits_record
from worker.framework.mixins import RedisMixin
from worker.test import get_last_processed_timestamp, create_scheduler_events


def generate_parser():
    """
    Creating the supporting arguments
    :return:
    """
    parser = argparse.ArgumentParser(description="Run your tests locally on an asset.")
    parser.add_argument("-a", "--asset_id", "--id", type=int, required=True, help="set asset_id")
    parser.add_argument("-s", "--start_timestamp", "--start", type=int, required=False, default=None, help="start timestamp")
    parser.add_argument("-e", "--end_timestamp", "--end", type=int, required=False, default=None, help="end timestamp")
    parser.add_argument("-i", "--timestep", "--step", type=int, required=False, default=60, help="trigger the lambda function once every step")
    parser.add_argument("-d", "--to_delete", "--delete", type=strtobool, required=False, default=False, help="to delete the state and data")
    return parser


class AppTestRun:
    def __init__(self, lambda_handler, collections):
        self.lambda_handler = lambda_handler
        self.collections = collections

        self.progress = None

        self.initialize()

    def initialize(self):
        parser = generate_parser()
        args = parser.parse_args()

        asset_id = args.asset_id
        start_timestamp = args.start_timestamp
        end_timestamp = args.end_timestamp
        step = args.timestep
        to_delete = args.to_delete

        app_redis = "corva/{0}.{1}".format(asset_id, constants.get("global.app-key"))

        if not start_timestamp:
            start_timestamp = get_one_wits_record(asset_id, timestamp_sort=+1)
        if not end_timestamp:
            end_timestamp = get_one_wits_record(asset_id, timestamp_sort=-1)
        if to_delete:
            RedisMixin.delete_states(pattern=f"{app_redis}*")
            for collection in self.collections:
                delete_collection_data_of_asset_id(asset_id, collection)

        start_timestamp = get_last_processed_timestamp(app_redis) or start_timestamp
        print(f"asset_id: {asset_id}, timestamp interval: [{start_timestamp}, {end_timestamp}]")

        events = create_scheduler_events(asset_id, start_timestamp, end_timestamp, step)

        self.progress = tqdm(events, ncols=150)

    def run(self):
        for e in self.progress:
            self.lambda_handler(e, None)
            self.progress.set_description(str(int(e[0][0]['schedule_start'] / 1000)))
